#include "BinaryTree.h"
#include <fstream>
#include <iomanip>
#include <iostream>
#include <stack>
using namespace std;

BinaryTree::BinaryTree(string filename) : root_(nullptr) {
	// Otvaramo fajl u 'in' rezimu rada (samo citanje)
	fstream inputFile(filename, ios::in);
	// Metoda 'peek()' dohvata sledeci karakter koji bi se procitao ali ga ne cita
	while (inputFile.peek() != EOF) {
		int number;
		// Fajl tokove koristimo isto kao i ulaz/izlaz tokove
		inputFile >> number;
		addValue(number);
	}
	// Na kraju je bitno da zatvorimo fajl kako bismo izbrisali sve res
	inputFile.close();
}

BinaryTree::BinaryTree(const BinaryTree & other) : root_(nullptr) {

	// Kopirajuci konstruktor treba da kopira sve cvorove i time kreira novo stablo
	// Koji obilazak ovde treba da se implementira?
	
	/* PREORDER
	kopiraj cvor
	ako postoji levi => obidji levi
	ako postoji desni => obidji desni
	*/

	TreeNode* node = other.root_;
	stack<TreeNode*> unvisitedNodes;
	while (!unvisitedNodes.empty() || node != nullptr) {
		if (node != nullptr) {
			// Cuvamo cvor na kojeg treba da se vratimo kasnije (pamtimo njegovu poziciju na steku)
			unvisitedNodes.push(node->getRightChild());
			// Vrsimo obradu cvora - kopiranje
			addValue(node->getValue());
			// Premestamo se na levo podstablo
			node = node->getLeftChild();
		}
		else {
			// Dohvatamo sa steka poziciju cvora na koji treba da se pozicioniramo 
			node = unvisitedNodes.top();
			// Skidamo sa steka jedan clan (njega smo prethodno sacuvali)
			unvisitedNodes.pop();
		}
	}
}

BinaryTree::BinaryTree(BinaryTree&& other) {
	root_ = other.root_;
	other.root_ = nullptr;
}

BinaryTree::~BinaryTree() {

	// Da bismo izbrisali cvor, treba da izbrisemo njegovo levo i desno podstablo, a tek onda da brisemo i sam cvor
	// Odavde zakljucujemo da je u pitanju postorder (levo podstablo - desno podstablo - brisanje cvora)

	/* POSTORDER
	ako postoji levi => obidji levi
	ako postoji desni => obidji desni
	obrisi cvor
	*/

	TreeNode* node = root_, *lastVisitedNode = nullptr;
	stack<TreeNode*> unvisitedNodes;
	while (!unvisitedNodes.empty() || node != nullptr) {
		if (node != nullptr) {
			// Cuvamo cvor na kojeg treba da se vratimo kasnije (pamtimo njegovu poziciju na steku)
			unvisitedNodes.push(node);
			// Premestamo se na levo podstablo
			node = node->getLeftChild();
		}
		else {
			// Uzimamo poslednje sacuvani cvor
			TreeNode* topNode = unvisitedNodes.top();
			// Ukoliko taj cvor ima desno dete i ukoliko to desno dete nismo obisli
			if (topNode->getRightChild() != nullptr && topNode->getRightChild() != lastVisitedNode) {
				// Premestamo se na desno dete
				node = topNode->getRightChild();
			}
			else {
				// U suprotnom smo sigurni da su i levo i desno podstablo obradjeni, te mozemo da brisemo tekuci cvor
				lastVisitedNode = topNode;
				delete topNode;
				unvisitedNodes.pop();
			}
		}
	}
}

// Metoda koja dodaje novu vrednost u stablo
void BinaryTree::addValue(int value){
	TreeNode* node = root_;
	// Ukoliko stablo ne postoji, dodaj korenu novi cvor na osnovu prosledjene vrednosti
	if (!root_) root_ = new TreeNode(value);
	else {
		// Ako stablo postoji, kreiraj nov cvor
		TreeNode* newNode = new TreeNode(value);
		// A zatim u petlji naci poziciju gde treba da smestimo taj novi cvor
		while (node) {
			// Ukoliko je nov cvor manji od trenutnog cvora u obilasku stabla, premesticemo se u levo podstablo
			if (node->getValue() > newNode->getValue()) {
				// Ukoliko levo podstablo ne postoji, nasli smo poziciju gde dodajemo nov cvor
				if (node->getLeftChild() == nullptr) {
					node->setLeftChild(newNode);
					break;
				}
				node = node->getLeftChild();
			}
			// Ukoliko je nov cvor veci od trenutnog cvora u obilasku stabla, premesticemo se u desno podstablo
			else if (node->getValue() < newNode->getValue()) {
				// Ukoliko desno podstablo ne postoji, nasli smo poziciju gde dodajemo nov cvor
				if (node->getRightChild() == nullptr) {
					node->setRightChild(newNode);
					break;
				}
				node = node->getRightChild();
			}
			// Ukoliko su cvorovi jednaki, izbrisacemo nov cvor iz dinamicke memorije, i zavrsicemo petlju (ovime se dodavanje cvora koji postoji u stablu ignorise)
			else {
				delete newNode;
				break;
			}
		}
	}
}

void BinaryTree::print() const {

	// Rekurzija ce da odradi posao, medjutim moze da dodje do stack overflowa ukoliko se previse ugnezdimo u rekurzivnim pozivima funkcije
	// Prilikom svakog poziva, na steku poziva se cuvaju informacije koje nam omogucavaju vracanje iz rekurzivnih poziva
	// U nekom trenutku, moze da se desi da se stek prepuni, i samim tim vise ne mozemo da pozivamo rekurzivne funkcije

	/* INORDER
	ako postoji levi => obidji levi
	ispisi cvor
	ako postoji desni => obidji desni
	*/

	// Iterativni postupak je iz tog razloga bolji

	// Koristi se inorder obilazak, koji nam odgovara za ispisivanje cvorova po rastucem redosledu
	// On podrazumeva obilazak i obradu levog podstabla, zatim obradu konkretnog cvora i na kraju obilazak i obradu desnog podstabla
	// Da bismo znali kako da se vratimo na cvor posle obilaska levog podstabla (da bismo mogli da ga obradimo i predjemo na obilazak desnog stabla), koristimo stek
	// Na taj nacin uvek cemo se vracati na cvor koji je poslednji stavljen na stek
	// Pre obilaska levog podstabla, na steku cuvamo poziciju tekuceg cvora,  a zatim obilazimo levo podstablo
	// Kad ne budemo vise mogli da obilazimo levo podstablo, sa steka skidamo poslednji cvor, obradjujemo ga (vrsimo ispis u ovom slucaju) i prelazimo na obilazak desnog podstabla

	TreeNode* node = root_;
	stack<TreeNode*> unvisitedNodes;
	while (!unvisitedNodes.empty() || node != nullptr)
	{
		if (node != nullptr)
		{
			// Cuvamo cvor na kojeg treba da se vratimo kasnije (pamtimo njegovu poziciju na steku)
			unvisitedNodes.push(node);
			// Premestamo se na levo podstablo
			node = node->getLeftChild();
		}
		else
		{
			// Dohvatamo sa steka poziciju cvora na koji treba da se pozicioniramo 
			node = unvisitedNodes.top();
			// Ispisujemo cvor
			cout << node->getValue() << endl;
			// Premestamo se na desno podstablo
			node = node->getRightChild();
			// Skidamo sa steka jedan clan (njega smo prethodno sacuvali)
			unvisitedNodes.pop();
		}
	}
}

void BinaryTree::print(string filename) const {
	fstream outputFile(filename, ios::out);
	TreeNode* node = root_;
	stack<TreeNode*> unvisitedNodes;
	while (!unvisitedNodes.empty() || node != nullptr)
	{
		if (node != nullptr)
		{
			unvisitedNodes.push(node);
			node = node->getLeftChild();
		}
		else
		{
			node = unvisitedNodes.top();
			// Na isti nacin upisujemo u izlazni fajl kao i na ekran
			// Sva formatiranja koja su vazila pre, vaze i ovde
			outputFile << setw(3) << setfill(' ') << node->getValue() << endl;
			node = node->getRightChild();
			unvisitedNodes.pop();
		}
	}
	outputFile.close();
}
